part of YIM;

class _YIMConversationManager {
  _YIMConversationManager._internal();

  static final _ = _YIMConversationManager._internal();

  factory _YIMConversationManager() {
    if (_YIMSDK.option == null) _YIMSDK.utils.yimLog("${_.runtimeType}:Please init First!");
    return _;
  }

  Future<void> queryRecentConversation({required void Function(List<YIMConversation>? conversations) callback}) async {
    if (!YIM.userManager.isLogin()) {
      try {
        callback.call(null);
      } catch (e) {}
      return;
    }
    final results = await _YIMSDK.db.query<YIMConversation>(
      _YIMTable.Conversation,
      onModel: (map) => YIMConversation.fromJson(map),
      finder: Finder(
        limit: 100,
        sortOrders: [SortOrder("lastMessage.time", false)],
      ),
    );
    if (results == null) {
      try {
        callback.call(null);
      } catch (e) {}
      return;
    }
    try {
      callback.call(results);
    } catch (e) {}
  }

  Future<void> clearUnreadCount({required YIMEnum_ConversationType conversationType, required String conversationID, bool deleteServer = true, void Function(bool isSuccess)? callback}) async {
    Future<void> clearUnreadCount_local() async {
      Filter? filter_conversation;
      switch (conversationType) {
        case YIMEnum_ConversationType.P2P:
          filter_conversation = Filter.and([
            Filter.or([
              Filter.and([Filter.equals("lastMessage.from", _YIMSDK.currentUser), Filter.equals("lastMessage.to", conversationID)]),
              Filter.and([Filter.equals("lastMessage.to", _YIMSDK.currentUser), Filter.equals("lastMessage.from", conversationID)]),
            ]),
            Filter.equals("lastMessage.conversationType", conversationType.rawValue),
          ]);
          break;
        case YIMEnum_ConversationType.Team:
          filter_conversation = Filter.and([
            Filter.equals("lastMessage.to", conversationID),
            Filter.equals("lastMessage.conversationType", conversationType.rawValue),
          ]);
          break;
      }
      final isSuccess_updateOne = await _YIMSDK.db.updateOne<YIMConversation>(
        _YIMTable.Conversation,
        (oldData) => oldData!..unRead = 0,
        onModel: (map) => YIMConversation.fromJson(map),
        finder: Finder(filter: filter_conversation),
      );
      if (!isSuccess_updateOne) {
        try {
          callback?.call(false);
        } catch (e) {}
        return;
      }
      try {
        YIM.listenerManager._callback_recentConversation?.call();
      } catch (e) {}
      try {
        callback?.call(true);
      } catch (e) {}
    }

    if (!YIM.userManager.isLogin()) {
      try {
        callback?.call(false);
      } catch (e) {}
      return;
    }
    if (deleteServer)
      _YIMSDK.server.emit("clearUnreadCount", data: {
        "conversationID": conversationID,
        "conversationType": conversationType.rawValue,
      }, ack: (isSuccess, results, code, msg) async {
        if (!isSuccess) {
          try {
            callback?.call(false);
          } catch (e) {}
          return;
        }
        await clearUnreadCount_local();
      });
    else
      await clearUnreadCount_local();
  }

  Future<void> clearAllUnreadCount({bool deleteServer = true, void Function(bool isSuccess)? callback}) async {
    Future<void> clearAllUnreadCount_local() async {
      final isSuccess_updateAll = await _YIMSDK.db.updateAll(
        _YIMTable.Conversation,
        () => {"unRead": 0},
      );
      if (!isSuccess_updateAll) {
        try {
          callback?.call(false);
        } catch (e) {}
        return;
      }
      try {
        YIM.listenerManager._callback_recentConversation?.call();
      } catch (e) {}
      try {
        callback?.call(true);
      } catch (e) {}
    }

    if (!YIM.userManager.isLogin()) {
      try {
        callback?.call(false);
      } catch (e) {}
      return;
    }
    if (deleteServer)
      _YIMSDK.server.emit("clearAllUnreadCount", ack: (isSuccess, results, code, msg) async {
        if (!isSuccess) {
          try {
            callback?.call(false);
          } catch (e) {}
          return;
        }
        await clearAllUnreadCount_local();
      });
    else
      await clearAllUnreadCount_local();
  }

  Future<void> delete({required YIMEnum_ConversationType conversationType, required String conversationID, bool deleteServer = true, void Function(bool isSuccess)? callback}) async {
    Future<void> delete_local() async {
      Filter? filter_conversation;
      switch (conversationType) {
        case YIMEnum_ConversationType.P2P:
          filter_conversation = Filter.and([
            Filter.or([
              Filter.and([Filter.equals("lastMessage.from", _YIMSDK.currentUser), Filter.equals("lastMessage.to", conversationID)]),
              Filter.and([Filter.equals("lastMessage.to", _YIMSDK.currentUser), Filter.equals("lastMessage.from", conversationID)]),
            ]),
            Filter.equals("lastMessage.conversationType", conversationType.rawValue),
          ]);
          break;
        case YIMEnum_ConversationType.Team:
          filter_conversation = Filter.and([
            Filter.equals("lastMessage.to", conversationID),
            Filter.equals("lastMessage.conversationType", conversationType.rawValue),
          ]);
          break;
      }
      final isSuccess_delete = await _YIMSDK.db.delete(
        _YIMTable.Conversation,
        finder: Finder(filter: filter_conversation),
      );
      if (!isSuccess_delete) {
        try {
          callback?.call(false);
        } catch (e) {}
        return;
      }
      try {
        YIM.listenerManager._callback_recentConversation?.call();
      } catch (e) {}
      try {
        callback?.call(true);
      } catch (e) {}
    }

    if (!YIM.userManager.isLogin()) {
      try {
        callback?.call(false);
      } catch (e) {}
      return;
    }
    if (deleteServer)
      _YIMSDK.server.emit("deleteConversation", data: {
        "conversationID": conversationID,
        "conversationType": conversationType.rawValue,
      }, ack: (isSuccess, results, code, msg) async {
        if (!isSuccess) {
          try {
            callback?.call(false);
          } catch (e) {}
          return;
        }
        await delete_local();
      });
    else
      await delete_local();
  }

  Future<bool> _updateLastMessage(YIMEnum_ConversationType conversationType, String conversationID, {_YIMEnum_ConversatypeUnRead unRead = _YIMEnum_ConversatypeUnRead.Nothing}) async {
    _YIMTable? tableName;
    Filter? filter_message;
    switch (conversationType) {
      case YIMEnum_ConversationType.P2P:
        tableName = _YIMTable.Message_P2P;
        filter_message = Filter.and([
          Filter.or([
            Filter.and([Filter.equals("from", _YIMSDK.currentUser), Filter.equals("to", conversationID)]),
            Filter.and([Filter.equals("to", _YIMSDK.currentUser), Filter.equals("from", conversationID)]),
          ]),
          Filter.equals("conversationType", conversationType.rawValue),
        ]);
        break;
      case YIMEnum_ConversationType.Team:
        tableName = _YIMTable.Message_Team;
        filter_message = Filter.and([
          Filter.equals("to", conversationID),
          Filter.equals("conversationType", conversationType.rawValue),
        ]);
        break;
    }
    final results = await _YIMSDK.db.query<YIMMessage>(
      tableName!,
      onModel: (map) => YIMMessage.fromJson(map),
      finder: Finder(
        limit: 1,
        sortOrders: [SortOrder("time", false)],
        filter: filter_message,
      ),
    );
    if (results == null) return false;
    if (results.isEmpty) return false;
    Filter? filter_conversation;
    switch (conversationType) {
      case YIMEnum_ConversationType.P2P:
        filter_conversation = Filter.and([
          Filter.or([
            Filter.and([Filter.equals("lastMessage.from", _YIMSDK.currentUser), Filter.equals("lastMessage.to", conversationID)]),
            Filter.and([Filter.equals("lastMessage.to", _YIMSDK.currentUser), Filter.equals("lastMessage.from", conversationID)]),
          ]),
          Filter.equals("lastMessage.conversationType", conversationType.rawValue),
        ]);
        break;
      case YIMEnum_ConversationType.Team:
        filter_conversation = Filter.and([
          Filter.equals("lastMessage.to", conversationID),
          Filter.equals("lastMessage.conversationType", conversationType.rawValue),
        ]);
        break;
    }
    final isSuccess_updateOrInsert = await _YIMSDK.db.updateOrInsert<YIMConversation>(
      _YIMTable.Conversation,
      (oldData) {
        final conversation = oldData ?? YIMConversation();
        if (results.first.from != _YIMSDK.currentUser)
          switch (unRead) {
            case _YIMEnum_ConversatypeUnRead.Plus:
              conversation.unRead++;
              break;
            case _YIMEnum_ConversatypeUnRead.Minus:
              if (conversation.unRead > 0) conversation.unRead--;
              break;
            case _YIMEnum_ConversatypeUnRead.Zero:
              conversation.unRead = 0;
              break;
          }
        conversation.lastMessage = results.first;
        return conversation;
      },
      finder: Finder(filter: filter_conversation),
      onModel: (map) => YIMConversation.fromJson(map),
    );
    if (!isSuccess_updateOrInsert) return false;
    return true;
  }
}
