part of YIM;

// Conversation：lastMessage unRead
// Message_P2P：id content from to conversationType messageType messageState time localExt isRevoke
// Message_Team：id content from to conversationType messageType messageState time localExt isRevoke
enum _YIMTable { Conversation, Message_P2P, Message_Team }

class _YIMDB {
  Map<String, Database> _databaseClients = {};

  Future<Database?> _getDatabaseClient(_YIMTable tableName) async {
    if (_YIMSDK.option == null) {
      _YIMSDK.utils.yimLog("DB: Please init First!");
      return null;
    }
    if (!YIM.userManager.isLogin()) {
      _YIMSDK.utils.yimLog("DB: Please login first!");
      return null;
    }
    final dbName = tableName.name;
    if (!_databaseClients.containsKey(dbName)) {
      final path = "/yim/${_YIMSDK.option!.appKey}/${_YIMSDK.currentUser}/${dbName}.db";
      _databaseClients[dbName] = kIsWeb ? await databaseFactoryWeb.openDatabase(path) : await databaseFactoryIo.openDatabase("${(await getApplicationDocumentsDirectory()).path}${path}");
    }
    return _databaseClients[dbName]!;
  }

  StoreRef<int, Map<String, Object?>> _getDB() => intMapStoreFactory.store(null);

  Future<int?> count(_YIMTable tableName, {Filter? filter}) async {
    final databaseClient = await _getDatabaseClient(tableName);
    if (databaseClient == null) return null;
    try {
      return await _getDB().count(databaseClient, filter: filter);
    } catch (e) {
      _YIMSDK.utils.yimLog("DB_c: ${e}");
      return null;
    }
  }

  Future<List<T>?> query<T extends Object>(_YIMTable tableName, {Finder? finder, required onModel(map)?}) async {
    final databaseClient = await _getDatabaseClient(tableName);
    if (databaseClient == null) return null;
    try {
      final List<dynamic> results = await _getDB().find(databaseClient, finder: finder);
      if (results.isEmpty) return [];
      if (onModel == null)
        return results as List<T>;
      else
        return results.map<T>((e) => onModel(e.value) as T).toList();
    } catch (e) {
      _YIMSDK.utils.yimLog("DB_q: ${e}");
      return null;
    }
  }

  Future<bool> insert(_YIMTable tableName, {Object? data, List<Object>? datas}) async {
    if (data == null && datas == null && datas!.isEmpty) {
      _YIMSDK.utils.yimLog("DB_i: No Data");
      return false;
    }
    final databaseClient = await _getDatabaseClient(tableName);
    if (databaseClient == null) return false;
    try {
      await _getDB().addAll(databaseClient, (datas ?? [data!]).map<Map<String, Object?>>((e) => json.decode(_YIMSDK.utils.toJson(e))).toList());
      return true;
    } catch (e) {
      _YIMSDK.utils.yimLog("DB_i: ${e}");
      return false;
    }
  }

  // 用于修改查询结果仅存在1条的数据
  Future<bool> updateOne<T extends Object>(_YIMTable tableName, T Function(T? oldData) onQueryFinish, {Finder? finder, required onModel(map)?}) async {
    final databaseClient = await _getDatabaseClient(tableName);
    if (databaseClient == null) return false;
    try {
      final results = await query<T>(tableName, finder: finder, onModel: onModel);
      if (results == null) return false;
      if (results.isNotEmpty) await _getDB().update(databaseClient, json.decode(_YIMSDK.utils.toJson(onQueryFinish(results.first))), finder: finder);
      return true;
    } catch (e) {
      _YIMSDK.utils.yimLog("DB_uo: ${e}");
      return false;
    }
  }

  // 用于修改查询结果大于等于0条的数据，批量修改指定字段
  Future<bool> updateAll(_YIMTable tableName, Map<String, dynamic> Function() onQueryFinish, {Finder? finder}) async {
    final databaseClient = await _getDatabaseClient(tableName);
    if (databaseClient == null) return false;
    try {
      await _getDB().update(databaseClient, json.decode(_YIMSDK.utils.toJson(onQueryFinish())), finder: finder);
      return true;
    } catch (e) {
      _YIMSDK.utils.yimLog("DB_ua: ${e}");
      return false;
    }
  }

  Future<bool> delete(_YIMTable tableName, {Finder? finder}) async {
    final databaseClient = await _getDatabaseClient(tableName);
    if (databaseClient == null) return false;
    try {
      await _getDB().delete(databaseClient, finder: finder);
      return true;
    } catch (e) {
      _YIMSDK.utils.yimLog("DB_d:${e}");
      return false;
    }
  }

  Future<bool> updateOrInsert<T extends Object>(_YIMTable tableName, T Function(T? data) onQueryFinish, {Finder? finder, required onModel(map)?}) async {
    try {
      final results = await query<T>(tableName, finder: finder, onModel: onModel);
      try {
        if (results == null) return false;
        if (results.isEmpty)
          await insert(tableName, data: onQueryFinish(null));
        else
          await updateOne<T>(tableName, (oldData) => onQueryFinish(oldData), finder: finder, onModel: onModel);
        return true;
      } catch (e) {
        _YIMSDK.utils.yimLog("DB_uoi:${e}");
        return false;
      }
    } catch (e) {
      _YIMSDK.utils.yimLog("DB_uoi:${e}");
      return false;
    }
  }

  void dispose() {
    _databaseClients.values.forEach((element) => element.close());
    _databaseClients.clear();
  }
}
