part of YIM;

class _YIMServer {
  void Function(_YIMResult<YIMEnum_LoginResult> result)? callback_login;
  Socket? socket;
  Timer? timer_checkConnect;
  Timer? timer_login;
  var isLoginDone = false;
  String? currentToken;
  bool isReconnecting = false;

  void init(String user, String token, void Function(_YIMResult<YIMEnum_LoginResult> result)? callback) {
    isLoginDone = false;
    isReconnecting = false;
    this.callback_login = callback;
    _YIMSDK.currentUser = user;
    currentToken = token;
    dispose(false);
    connect(false);
    timer_login = Timer.periodic(Duration(milliseconds: _YIMSDK.config.socketServerConnectTimeout), (timer) {
      onServer("loginFailed", {"msg": "time out"});
    });
  }

  void connect(bool isReconnect) {
    String getDevice() {
      if (kIsWeb) return "web";
      if (Platform.isWindows || Platform.isMacOS || Platform.isLinux) return "pc";
      if (Platform.isAndroid) return "android";
      if (Platform.isIOS) return "ios";
      return "";
    }

    final params = {
      "accid": _YIMSDK.currentUser,
      "token": currentToken,
      "appKey": _YIMSDK.option?.appKey ?? "",
      "device": getDevice(),
      "isReconnect": isReconnect ? "1" : "0",
    };
    final option = OptionBuilder();
    option.setTimeout(_YIMSDK.config.socketServerConnectTimeout);
    option.setTransports(['websocket']);
    option.disableAutoConnect();
    option.disableReconnection();
    option.setQuery(params);
    socket = io(_YIMSDK.config.server, option.build());
    socket!.onConnect((data) => _YIMSDK.utils.yimLog("onConnect"));
    socket!.onReconnect((data) => _YIMSDK.utils.yimLog("onReconnect"));
    socket!.onConnecting((data) => _YIMSDK.utils.yimLog("onConnecting"));
    socket!.onConnectTimeout((data) => _YIMSDK.utils.yimLog("onConnectTimeout"));
    socket!.onConnectError((data) => _YIMSDK.utils.yimLog("onConnectError"));
    socket!.onDisconnect((data) => _YIMSDK.utils.yimLog("onDisconnect"));
    socket!.on('server', (data_server) async => await onServer(data_server["event"], data_server["data"]));
    socket!.connect();
  }

  void emit(String event, {Map<String, Object?>? data, void Function(bool isSuccess, Map<String, dynamic>? results, int code, String msg)? ack}) {
    _YIMSDK.utils.yimLog("[emit_request]: \nevent: $event\ndata: ${_YIMSDK.utils.toJson(data ?? {})}");
    var isFinish = false;
    try {
      if (!YIM.userManager.isLogin()) {
        if (isFinish) return;
        isFinish = true;
        _YIMSDK.utils.yimLog("[emit_cancel]: \nevent: $event\ndata: ${_YIMSDK.utils.toJson(data ?? {})}");
        try {
          ack?.call(false, null, 0, "Please login first!");
        } catch (e) {}
        return;
      }
      Future.delayed(Duration(milliseconds: _YIMSDK.config.socketServerConnectTimeout), () {
        if (isFinish) return;
        isFinish = true;
        _YIMSDK.utils.yimLog("[emit_timeout]: \nevent: $event\ndata: ${_YIMSDK.utils.toJson(data ?? {})}");
        try {
          ack?.call(false, null, 0, "[$event] time out");
        } catch (e) {}
      });
      socket?.emitWithAck("client", {
        "event": event,
        "data": data,
      }, ack: (data) {
        try {
          Map<String, dynamic>? results = data;
          if (results == null) {
            if (isFinish) return;
            isFinish = true;
            _YIMSDK.utils.yimLog("[emit_response_null]: \nevent: $event\ndata: ${_YIMSDK.utils.toJson(data ?? {})}");
            try {
              ack?.call(false, null, 0, "null response");
            } catch (e) {}
            return;
          }
          final code = results["code"] ?? 0;
          final isSuccess = (code == 200);
          final msg = results["msg"] ?? "";
          results.remove("code");
          results.remove("msg");
          if (isFinish) return;
          isFinish = true;
          _YIMSDK.utils.yimLog("[emit_response_success]: \nevent: $event\ndata: ${_YIMSDK.utils.toJson(data ?? {})}");
          try {
            ack?.call(isSuccess, results["data"], code, msg);
          } catch (e) {}
        } catch (e) {
          if (isFinish) return;
          isFinish = true;
          _YIMSDK.utils.yimLog("[emit_exception_1]: \nevent: $event\ndata: ${_YIMSDK.utils.toJson(data ?? {})}");
          try {
            ack?.call(false, null, 0, "Exception: ${e.toString()}");
          } catch (e) {}
        }
      });
    } catch (e) {
      if (isFinish) return;
      isFinish = true;
      _YIMSDK.utils.yimLog("[emit_exception_2]: \nevent: $event\ndata: ${_YIMSDK.utils.toJson(data ?? {})}");
      try {
        ack?.call(false, null, 0, "Exception: ${e.toString()}");
      } catch (e) {}
    }
  }

  void dispose(bool isReconnect) {
    if (isReconnect) {
      socket?.clearListeners();
      socket?.dispose();
      socket = null;
    } else {
      // currentToken = null;
      timer_checkConnect?.cancel();
      timer_checkConnect = null;
      timer_login?.cancel();
      timer_login = null;
      socket?.clearListeners();
      socket?.dispose();
      socket = null;
    }
  }

  Future<void> onServer(String event, Map<String, dynamic> data) async {
    Future<void> doMessages(List<YIMMessage> messages, bool isNewMessage, {void Function(bool isSuccess)? callback}) async {
      for (var i = 0; i < messages.length; i++) {
        final message = messages[i];
        switch (message.conversationType) {
          case YIMEnum_ConversationType.P2P:
            final count = await _YIMSDK.db.count(
              _YIMTable.Message_P2P,
              filter: Filter.equals("id", message.id),
            );
            if (count == null) {
              if (i == messages.length - 1) callback?.call(false);
              return;
            }
            if (count > 0) {
              if (i == messages.length - 1) callback?.call(false);
              return;
            }
            final isSuccess_insert = await _YIMSDK.db.insert(
              _YIMTable.Message_P2P,
              data: message,
            );
            if (!isSuccess_insert) {
              if (i == messages.length - 1) callback?.call(false);
              return;
            }
            final isSuccess_updateLastMessage = await YIM.conversationManager._updateLastMessage(
              message.conversationType,
              message.from == _YIMSDK.currentUser ? message.to : message.from,
              unRead: _YIMEnum_ConversatypeUnRead.Plus,
            );
            if (!isSuccess_updateLastMessage) {
              if (i == messages.length - 1) callback?.call(false);
              return;
            }
            try {
              YIM.listenerManager._callback_recentConversation?.call();
            } catch (e) {}
            if (isNewMessage)
              try {
                YIM.listenerManager._callback_message?.call(messages);
              } catch (e) {}
            if (i == messages.length - 1) callback?.call(true);
            break;
          case YIMEnum_ConversationType.Team:
            final count = await _YIMSDK.db.count(
              _YIMTable.Message_Team,
              filter: Filter.equals("id", message.id),
            );
            if (count == null) {
              if (i == messages.length - 1) callback?.call(false);
              return;
            }
            if (count > 0) {
              if (i == messages.length - 1) callback?.call(false);
              return;
            }
            final isSuccess_insert = await _YIMSDK.db.insert(
              _YIMTable.Message_Team,
              data: message,
            );
            if (!isSuccess_insert) {
              if (i == messages.length - 1) callback?.call(false);
              return;
            }
            final isSuccess_updateLastMessage = await YIM.conversationManager._updateLastMessage(
              message.conversationType,
              message.to,
              unRead: _YIMEnum_ConversatypeUnRead.Plus,
            );
            if (!isSuccess_updateLastMessage) {
              if (i == messages.length - 1) callback?.call(false);
              return;
            }
            try {
              YIM.listenerManager._callback_recentConversation?.call();
            } catch (e) {}
            if (isNewMessage)
              try {
                YIM.listenerManager._callback_message?.call(messages);
              } catch (e) {}
            if (i == messages.length - 1) callback?.call(true);
            break;
          default:
            break;
        }
      }
    }

    switch (event) {
      case "loginFailed":
        if (isLoginDone) return;
        isLoginDone = true;
        YIM.userManager.logout();
        try {
          callback_login?.call(_YIMResult<YIMEnum_LoginResult>(code: YIMEnum_LoginResult.Failed, msg: data["msg"]));
        } catch (e) {}
        callback_login = null;
        break;
      case "loginSuccess":
        if (isLoginDone) return;
        isLoginDone = true;
        timer_checkConnect?.cancel();
        timer_checkConnect = null;
        timer_login?.cancel();
        timer_login = null;
        timer_checkConnect = Timer.periodic(Duration(seconds: 5), (timer) {
          if (!YIM.userManager.isLogin()) {
            isReconnecting = true;
            try {
              YIM.listenerManager._callback_status_online?.call(_YIMResult<YIMEnum_OnlineState>(code: YIMEnum_OnlineState.Reconnecting, msg: "Reconnecting"));
            } catch (e) {}
            dispose(true);
            Future.delayed(Duration(seconds: 2), () => connect(true));
          } else {
            if (isReconnecting) {
              isReconnecting = false;
              try {
                YIM.listenerManager._callback_status_online?.call(_YIMResult<YIMEnum_OnlineState>(code: YIMEnum_OnlineState.Reconnected, msg: "Reconnected"));
              } catch (e) {}
            }
            _YIMSDK.server.emit("refreshOnline");
          }
        });
        //清理旧消息
        final count = await _YIMSDK.db.count(
          _YIMTable.Message_Team,
        );
        if (count != null && count > _YIMSDK.config.maxTeamMessageCount)
          _YIMSDK.db.delete(
            _YIMTable.Message_Team,
            finder: Finder(
              limit: _YIMSDK.config.maxTeamMessageCount ~/ 2,
              sortOrders: [SortOrder("time", false)],
            ),
          );
        //同步云端会话
        emit("queryConversations", ack: (isSuccess, results, code, msg) {
          if (isSuccess) {
            List<Map<String, dynamic>> conversations_server = (json.decode(results?["conversations"]?.toString() ?? "[]") as List<dynamic>).map((m) => m as Map<String, dynamic>).toList();
            if (conversations_server.isEmpty) return;
            YIM.conversationManager.queryRecentConversation(
              callback: (conversations) {
                //过滤本地已经存在的云端会话
                conversations?.forEach((conversation) {
                  switch (conversation.lastMessage!.conversationType) {
                    case YIMEnum_ConversationType.P2P:
                      final conversationID = conversation.lastMessage!.from == _YIMSDK.currentUser ? conversation.lastMessage!.to : conversation.lastMessage!.from;
                      for (var i = conversations_server.length - 1; i >= 0; i--) {
                        if (conversations_server[i]["conversationID"] == conversationID) conversations_server.removeAt(i);
                      }
                      break;
                    case YIMEnum_ConversationType.Team:
                      final conversationID = conversation.lastMessage!.to;
                      for (var i = conversations_server.length - 1; i >= 0; i--) {
                        if (conversations_server[i]["conversationID"] == conversationID) conversations_server.removeAt(i);
                      }
                      break;
                    default:
                      break;
                  }
                });
                //将本地没有的云端会话，把消息同步下来
                for (final conversation in conversations_server) {
                  YIM.messageManager.queryServerMessages(
                    conversationType: YIMEnum_ConversationType.from(conversation["conversationType"].toString()),
                    conversationID: conversation["conversationID"].toString(),
                    limit: _YIMSDK.config.roamMessageSize,
                    callback: (isSuccess, messages) {
                      if (isSuccess) {
                        doMessages(messages!, false, callback: (isSuccess) async {
                          if (!isSuccess) return;
                          Filter? filter;
                          switch (YIMEnum_ConversationType.from(conversation["conversationType"].toString())) {
                            case YIMEnum_ConversationType.P2P:
                              filter = Filter.and([
                                Filter.or([
                                  Filter.and([Filter.equals("lastMessage.from", _YIMSDK.currentUser), Filter.equals("lastMessage.to", conversation["conversationID"])]),
                                  Filter.and([Filter.equals("lastMessage.to", _YIMSDK.currentUser), Filter.equals("lastMessage.from", conversation["conversationID"])]),
                                ]),
                              ]);
                              break;
                            case YIMEnum_ConversationType.Team:
                              filter = Filter.and([
                                Filter.equals("lastMessage.to", conversation["conversationID"]),
                              ]);
                              break;
                            case YIMEnum_ConversationType.ChatRoom:
                              break;
                            default:
                              break;
                          }
                          final isSuccess_updateOne = await _YIMSDK.db.updateOne<YIMConversation>(
                            _YIMTable.Conversation,
                            (oldData) => oldData!..unRead = int.parse(conversation["unReadCount"].toString()),
                            onModel: (map) => YIMConversation.fromJson(map),
                            finder: Finder(filter: filter),
                          );
                          if (!isSuccess_updateOne) return;
                          try {
                            YIM.listenerManager._callback_recentConversation?.call();
                          } catch (e) {}
                        });
                      }
                    },
                  );
                }
              },
            );
          }
        });
        try {
          callback_login?.call(_YIMResult<YIMEnum_LoginResult>(code: YIMEnum_LoginResult.Success, msg: "success"));
        } catch (e) {}
        callback_login = null;
        break;
      case "kick":
        YIM.userManager.logout();
        try {
          YIM.listenerManager._callback_status_online?.call(_YIMResult<YIMEnum_OnlineState>(code: YIMEnum_OnlineState.Kick, msg: "Kicked by ${data["device"]} device"));
        } catch (e) {}
        break;
      case "message":
        final messages = (data["messages"] as List<dynamic>).map((element) => YIMMessage.fromJson(element)).toList();
        doMessages(messages, true);
        break;
      case "broadcast":
        try {
          YIM.listenerManager._callback_broadcast?.call(_YIMBroadcastMessage.fromJson(data["message"]));
        } catch (e) {}
        break;
      case "notification":
        try {
          YIM.listenerManager._callback_notification?.call(_YIMNotificationMessage.fromJson(data["message"]));
        } catch (e) {}
        break;
      case "revoke":
        for (final element in (data["messages"] as List<dynamic>)) {
          final message = YIMRevokeMessage.fromJson(element);
          final isSuccess_markConversationRevoke = await YIM.messageManager._markConversationRevoke(
            message.conversationType,
            message.messageID,
          );
          if (!isSuccess_markConversationRevoke) return;
          final isSuccess_updateLastMessage = await YIM.conversationManager._updateLastMessage(
            message.conversationType,
            message.conversationID,
            // unRead: _YIMEnum_ConversatypeUnRead.Minus,
          );
          if (!isSuccess_updateLastMessage) return;
          try {
            YIM.listenerManager._callback_recentConversation?.call();
          } catch (e) {}
          try {
            YIM.listenerManager._callback_revoke?.call(message);
          } catch (e) {}
        }
        break;
      case "receipt":
        List<String> conversationIDs = (data["conversationIDs"] as List).map((element) => element.toString()).toList();
        for (final conversationID in conversationIDs) {
          final isSuccess_markConversationRead = await YIM.messageManager._markConversationRead(
            conversationID,
            false,
          );
          if (!isSuccess_markConversationRead) return;
          final isSuccess_updateLastMessage = await YIM.conversationManager._updateLastMessage(
            YIMEnum_ConversationType.P2P,
            conversationID,
          );
          if (!isSuccess_updateLastMessage) return;
          try {
            YIM.listenerManager._callback_recentConversation?.call();
          } catch (e) {}
          try {
            YIM.listenerManager._callback_status_msg_receipt?.call(conversationID);
          } catch (e) {}
        }
        break;
    }
  }
}
